#ifndef __LINEAR_SCAN_H_
#define __LINEAR_SCAN_H_

#include "codegen/MachineCodePass.hpp"
#include "codegen/MachineOperand.hpp"

#include <map>
#include <set>
#include <string>
#include <string_view>
#include <vector>
#include <stack>

class MachineUnit;
class MachineFunction;

class LinearScan : public MachineCodePass {
public:
    struct Interval {
        std::string vreg_name;
        int begin, end;
        bool spilled{};
        IREGISTER phy_reg{};
        int offset{};
    };

    struct IntervalBeginCmper {
        bool operator() (const Interval *a, const Interval *b) const { return a->begin < b->begin; }
    };
    struct IntervalEndCmper {
        bool operator() (const Interval *a, const Interval *b) const { return a->end < b->end; }
    };

    std::vector<int> init_free_regs = {
        T0, T1, T2, T3, T4, T5, T6,
        S1, S2, S3, S4, S5, S6, S7, S8, S9, S10, S11,
    };
    std::vector<int> free_regs;

    std::vector<Interval *> intervals;
    std::multiset<Interval *, IntervalEndCmper> active;
    
    std::map<std::string, Interval *> reg_allocation;
    std::map<std::string, std::set<int>> caller_saved_regs;
    std::map<std::string, std::set<int>> callee_saved_regs;
        
    void pass(MachineUnit *mUnit) override;
    void computeLiveIntervals(MachineFunction *mFunc);
    void linearScanRegisterAllocation(MachineFunction *mFunc);
    void expireOldIntervals(Interval *i);
    void spillInterval(Interval *i);
    void updateUsedPhysicalRegs(const std::string &func_name, IREGISTER reg);
    void rewrite(MachineFunction *mFunc);
};

#endif