#ifndef __MACHINE_OPERAND_H_
#define __MACHINE_OPERAND_H_

#include <memory>
#include <ostream>

class MachineInstruction;

class MachineOperand {
public:
    enum MachineOperandType{ IMM, VREG, REG, LABEL };
    
    MachineInstruction *parent;
    MachineOperandType type;
    int val;
    int reg_no;
    std::string label;
    std::string vreg_name;
    bool param{false};
    int paramNo;
    bool fpu{false};
    float fval;

    
    MachineOperand() = delete;
    MachineOperand(MachineOperandType type, int val, bool fpu = false); // IMM_I, I_REG, F_REG
    MachineOperand(MachineOperandType type, std::string label_p); // VREG, LABEL
    // MachineOperand(int type_p, float fval_p);

    void print_reg(std::ostream &os) const;
    void output(std::ostream &os) const;
};

enum IREGISTER {
    ZERO = 0,
    RA,
    SP,
    GP,
    TP,
    T0, T1, T2,
    FP,
    S1,
    A0, A1, A2, A3, A4, A5, A6, A7,
    S2, S3, S4, S5, S6, S7, S8, S9, S10, S11,
    T3, T4, T5, T6,
};

#endif