//
// Created by q1w2e3r4 on 23-7-9.
//

#include "BasicBlock.h"
#include "IRInstruction.h"
#include "Utils.h"
#include "codegen/AsmBuilder.hpp"
#include "codegen/MachineBlock.hpp"
#include "codegen/MachineInstruction.hpp"
#include "codegen/MachineOperand.hpp"
#include <memory>
#include "IRInstruction.h"

using namespace std;

BasicBlock::BasicBlock(std::string& label,Function * func) {
    this->label = Utils::getNewLabel(label);
    this->func_belong = func;
    this->code = "";
}

void BasicBlock::appendCode(Instruction* instruction,std::string &code) {
    this->local_instr.push_back(instruction);
    this->code += "  " + code + "\n";
}

void BasicBlock::insertCodeAtFront(Instruction *instruction, string &code) {
    auto& v = this->local_instr;
    v.insert(v.begin(),instruction);
    this->code = "  " + code + "\n" + this->code;
}

void BasicBlock::Emit() {
    cout << this->label + ":            ";
    if(this->pred.empty()){
        cout << ";No predecessor!!";
    }
    else{
        cout << ";preds: ";
        for(BasicBlock* block : pred) {
            cout << block->label << ',';
        }
    }
    cout << endl;
    for(auto instr : this->local_instr){
        cout << "  ";
        instr->outPut();
    }
}

void BasicBlock::codegen(AsmBuilder *builder,
                         std::map<std::string, int> &offset_table,
                         std::map<std::string, int> &size_table,
                         int framesize,
                         vector<Type*> &arguments,
                         bool is_entry) {
    auto cur_func = builder->getFunction();
    auto cur_block = new MachineBlock(cur_func, this);
    mBlock = cur_block, cur_block->bb = this;
    builder->setBlock(cur_block);
    if (is_entry) {
        auto addi = new BinaryMInstruction(cur_block,
                                                BinaryMInstruction::ADDI,
                                                new MachineOperand(MachineOperand::REG, IREGISTER::SP), //x2(sp)
                                                new MachineOperand(MachineOperand::REG, IREGISTER::SP),//x2(sp)
                                                new MachineOperand(MachineOperand::IMM, -framesize));
        cur_block->insertInst(addi);
        auto sd1 = new StoreMInstruction(cur_block,
                                                StoreMInstruction::SD,
                                                new MachineOperand(MachineOperand(MachineOperand::REG, IREGISTER::RA)),
                                                new MachineOperand(MachineOperand::IMM, framesize - 8),
                                                new MachineOperand(MachineOperand::REG, IREGISTER::SP));
        cur_block->insertInst(sd1);
        auto sd2 = new StoreMInstruction(cur_block,
                                                StoreMInstruction::SD,
                                                new MachineOperand(MachineOperand(MachineOperand::REG, IREGISTER::FP)),
                                                new MachineOperand(MachineOperand::IMM, framesize - 16),
                                                new MachineOperand(MachineOperand::REG, IREGISTER::SP));
        cur_block->insertInst(sd2);
        auto adds0 = new BinaryMInstruction(cur_block,
                                           BinaryMInstruction::ADDI,
                                           new MachineOperand(MachineOperand::REG, IREGISTER::FP), //x2(sp)
                                           new MachineOperand(MachineOperand::REG, IREGISTER::SP),//x2(sp)
                                           new MachineOperand(MachineOperand::IMM, framesize));
        cur_block->insertInst(adds0);

//        arguments
        for(int i=0; i<arguments.size(); i++){
            if(Type_Enum(arguments[i]->type) == INT32TYPE){
                auto addArgument = new BinaryMInstruction(cur_block,
                                                    BinaryMInstruction::ADDI,
                                                    new MachineOperand(MachineOperand::VREG, "%" + to_string(i)), //x2(sp)
                                                    new MachineOperand(MachineOperand::REG, IREGISTER::A0+i),//x2(sp)
                                                    new MachineOperand(MachineOperand::IMM, 0));
                cur_block->insertInst(addArgument);
            }
            else if(Type_Enum(arguments[i]->type) == POINTERTYPE){
                auto addArgument = new BinaryMInstruction(cur_block,
                                                          BinaryMInstruction::ADDI,
                                                          new MachineOperand(MachineOperand::VREG, "%" + to_string(i)),
                                                          new MachineOperand(MachineOperand::REG, IREGISTER::A0+i),
                                                          new MachineOperand(MachineOperand::IMM, 0));
                cur_block->insertInst(addArgument);
            }
        }

    }

    for (auto inst : local_instr) {
        inst->codegen(builder,offset_table, size_table, framesize);
    }

    cur_func->insert_block(cur_block);
}

