//
// Created by q1w2e3r4 on 23-7-17.
//

#include "Function.h"
#include "IRInstruction.h"
#include <queue>
Function::Function(string name, FuncType *funcType) {
    this->name = std::move(name);
    this->funcType = funcType;
}

void Function::appendBlock(BasicBlock *block) {
    this->block_list.push_back(block);
}


void Function::Emit() {
    cout << this->debugInfo;
    for(BasicBlock *block : block_list){



        block -> Emit();

        //TODO:为解决 error:expected opcode即空block的bug，但是暂时想不出根本解决的方法，用补空指令的方法代替
        //TODO：根除这些空block的方法是提前知道当前block后面还有没有stmt，如何做到呢？
        if (block->code == "") {
            if (this->funcType->retType->type == VOIDTYPE)
                cout << "  ret void\n";
            else
                cout << "  ret " + this->funcType->retType->getTypeName() + " 0\n";
        }
    }
    if (this->debugInfo.length() != 0) cout << "}" << endl;
}

void Function::PrintSymbol() {
    cout << this->name << " Symbols:" << endl;
    for(const auto& it:symbol_table){
        cout << it.first << ": " << it.second->symbolType->getSize() << endl;
    }
}

void Function::codegen(AsmBuilder *builder) {
        auto cur_unit = builder->getUnit();
        auto cur_func = new MachineFunction(cur_unit, this);
        builder->setFunction(cur_func);

//        z:offset is the size of the func?
        int offset = 0;
        std::map<std::string, int> offset_table;
        std::map<std::string, int> size_table;
        for (auto &[sym_name, sym] : symbol_table) {
            // std::cerr << sym_name << ' ' << sym->symbolType->getSize() << endl;
            offset_table[sym_name] = offset;
            int size = sym->symbolType->getSize();
            size_table[sym_name] = size;
            offset += size;
        }
        cur_func->framesize = offset + 16;
//        func->codegen call block->codegen where is the first block?
        // block_list[0]->codegen(builder, offset_table, size_table, offset + 16,this->funcType->arguments, true);

        // TODO: prepare for multi blocks
       for(int i=0;i<block_list.size();i++){
           block_list[i]->codegen(builder, offset_table, size_table, offset + 16, this->funcType->arguments, i == 0);
       }

//        auto cur_block = builder->getBlock();
//        auto ld1 = new LoadMInstruction(cur_block,
//                                                LoadMInstruction::LD,
//                                                new MachineOperand(MachineOperand::REG, IREGISTER::RA),
//                                                new MachineOperand(MachineOperand::IMM, offset + 16 - 8), // framesize - 8
//                                                new MachineOperand(MachineOperand::REG, IREGISTER::SP));
//        cur_block->insertInst(ld1);
//        auto ld2 = new LoadMInstruction(cur_block,
//                                                LoadMInstruction::LD,
//                                                new MachineOperand(MachineOperand::REG, IREGISTER::FP),
//                                                new MachineOperand(MachineOperand::IMM, offset), // framesize
//                                                new MachineOperand(MachineOperand::REG, IREGISTER::SP));
//        cur_block->insertInst(ld2);
//        auto addi = new BinaryMInstruction(cur_block,
//                                           BinaryMInstruction::ADDI,
//                                           new MachineOperand(MachineOperand::REG, IREGISTER::SP),
//                                           new MachineOperand(MachineOperand::REG, IREGISTER::SP),
//                                           new MachineOperand(MachineOperand::IMM, offset+16));
//        cur_block->insertInst(addi);
//        auto jr = new JrMInstruction(cur_block,
//                                     JrMInstruction::JR,
//                                     new MachineOperand(MachineOperand::REG, IREGISTER::RA));
//        cur_block->insertInst(jr);
//        IRInstruction::retFunc(builder, offset);

        cur_unit->insert_func(cur_func);
    }

static void dfs(BasicBlock *cur, std::vector<BasicBlock *> &result, std::set<BasicBlock *> &visited) {
    // std::cerr << cur->label << ' ';
    visited.insert(cur);
    for (auto bb : cur->succ)
        if (visited.find(bb) == visited.end()) dfs(bb, result, visited);
    result.push_back(cur);
}

std::vector<BasicBlock *> Function::getReversePostOrder() {
    std::vector<BasicBlock *> result{};
    std::set<BasicBlock *> visited{};
    // std::cerr << "traverse func \"" << this->name << "\" ";
    dfs(block_list[0], result, visited);
    // std::cerr << std::endl;
    return std::vector<BasicBlock *>(result.rbegin(), result.rend());
    /*返回本函数内基本块的拓扑排序
     * 此功能正确的前提是，函数至少有一个基本块，且第一个块一定为Entry块
     * */
    // vector<BasicBlock *> ret;
    // queue<BasicBlock *> q;
    // set<BasicBlock *> s;
    // if(block_list.size() == 0) return ret;
    // q.push(this->block_list[0]);
    // while(!q.empty()){
    //     BasicBlock * now = q.front(); q.pop();
    //     ret.push_back(now);
    //     s.insert(now);

    //     for(auto bb: now->succ){
    //         bool flag = true;
    //         for(auto it : bb->pred){
    //             if(s.count(it) == 0){
    //                 flag = false;
    //                 break;
    //             }
    //         }

    //         if(flag){
    //             q.push(bb);
    //         }
    //     }
    // }
    // return ret;
}
