//
// Created by q1w2e3r4 on 23-7-27.
//

#include "HIR-opt/Mem2RegPass.h"
#include "IRInstruction.h"
#include "HIR-opt/HOptUtils.h"
#include<map>
#include<set>
#include<stack>
using namespace std;

// insertPhi uses
map<ValueRef*,set<BasicBlock*>> defsites; // describe what blocks def this ref
map<BasicBlock*,set<ValueRef*>> placed; // describe what ref need to be phi in one block
Undef * undef = new Undef();
// renamePhi uses
map<ValueRef*,stack<ValueRef*>> st;// stack top = live var(we use it to replace symbol)

void Mem2RegPass::run() {
    for(auto& [name,func] : globalUnit->func_table){
        Utils::labelCounter = 0;
        insertPhi(func);
        if(func->dom_root != nullptr) {
            st.clear();
            renamePhi(func->dom_root);
        }
    }
}

void Mem2RegPass::insertPhi(Function *func) {
    for(auto block:func->block_list){ // cal defsites
        for(auto var:block->live_def){
            defsites[var].insert(block);
        }
    }

    for(auto& [var,defs]: defsites){
        bool is_int = ((Symbol*)var)->symbolType->type == INT32TYPE;
        while(!defs.empty()){
            BasicBlock* block = *(defs.begin()); defs.erase(defs.begin());
            auto DF = block->domNode->DF;
            for(auto df : DF){
                BasicBlock* bb = df->bb;
                if(!placed[bb].count(var)){
                    ValueRef* result;
                    if(is_int)
                        result = new Int_Var("%phi_");
                    else
                        result = new Float_Var("%phi_");

                    IRInstruction * phi = new PhiInstruction(var,result,bb->pred.size());
                    phi->block = bb;
                    string debugCode = result->get_Ref() + " = phi(" + to_string(bb->pred.size()) + ")";
                    bb->insertCodeAtFront(phi,debugCode);
                    placed[bb].insert(var);

                    if(!bb->live_def.count(var)){
                        defsites[var].insert(bb);
                    }
                }
            }
        }
    }
}

void Mem2RegPass::renamePhi(DomNode * root) {
    map<ValueRef*,int> counter; // mark how many times one var is defined(push into stack) in this block
    BasicBlock* nowBlock = root->bb;
    auto& instructions = nowBlock->local_instr;
    for(auto it = instructions.begin(); it != instructions.end();){
        if((*it)->instType == LOAD){ // use
            // load src -> dst,we want to delete this LOADinstr, so replace dst
            ValueRef* src = ((LoadInstruction*)(*it))->src;
            ValueRef* dst = ((LoadInstruction*)(*it))->dst;
            if(dynamic_cast<Symbol*>(src) -> is_global) { // we don't simplify global instr;
                ++it;
                continue;
            }
            ValueRef* replaced = st[src].top();
            Replace(dst,replaced);
            it = instructions.erase(it);
        }
        else if((*it)->instType == STORE) { // def
            ValueRef* src = ((StoreInstruction*)(*it))->src;
            ValueRef* dst = ((StoreInstruction*)(*it))->dst;
            // store src -> dst,  we record src as def to replace other uses
            st[dst].push(src); counter[dst]++;
            it = instructions.erase(it);
        }
        else if((*it)->instType == PHI){ //
            ValueRef* var = ((PhiInstruction*)(*it))->symbol;
            ValueRef* value = ((PhiInstruction*)(*it))->result;
            st[var].push(value); counter[var]++;
            ++it;
        }
        else{
            ++it;
        }
    }

    for(auto next: nowBlock->succ){ // rename phi node
        for(auto instr: next->local_instr){
            if(instr->instType != PHI) break;
            auto phi = (PhiInstruction*)instr;
            ValueRef* var = phi->symbol;
            if(st[var].empty()){
                phi->mp[nowBlock] = undef;
            }
            else {
                phi->mp[nowBlock] = st[var].top();
            }
        }
    }

    for(auto child: root->child){ // recursive-call
        renamePhi(child);
    }

    for(auto&[var,cnt]:counter){
        st[var].pop();
    }
}
