//
// Created by q1w2e3r4 on 23-7-10.
//

#include "LLVM.h"

bool Both_Const(ValueRef* lhs, ValueRef* rhs);
bool Both_Int(ValueRef* lhs, ValueRef* rhs);

ValueRef* BuildLT(Builder* builder, ValueRef* lhs, ValueRef* rhs);
ValueRef* BuildLE(Builder* builder, ValueRef* lhs, ValueRef* rhs);
ValueRef* BuildEquals(Builder* builder, ValueRef* lhs, ValueRef* rhs);
ValueRef* BuildUnequal(Builder* builder, ValueRef* lhs, ValueRef* rhs);

ValueRef* BuildXorTrue(Builder* builder,ValueRef* ref); // SysY语法里没有xor,唯一用到的是和true的异或

string zeroInitializer(Type *type);
string nonZeroInitializer(std::any values,Type* type);

ValueRef* BuildAdd(Builder* builder, ValueRef* lhs, ValueRef* rhs){
    if(Both_Const(lhs,rhs)){
        if(lhs->type == rhs->type && lhs->type == 0){
            int op1 = stoi(lhs->get_Ref());
            int op2 = stoi(rhs->get_Ref());
            return new Int_Const(op1 + op2);
        }
        else{
            float op1 = stof(lhs->get_Ref());
            float op2 = stof(rhs->get_Ref());
            return new Float_Const(op1 + op2);
        }
    }
    else if(Both_Int(lhs,rhs)){
        auto* ret = new Int_Var();
        string src1 = lhs->get_Ref(), src2 = rhs->get_Ref(), dest = ret->get_Ref();
        string code = dest + " = add i32 "+ src1 + ", " + src2;

        IRInstruction * instr = new BinaryInstruction(ret,lhs,rhs,ADD);
        builder->AppendCode(instr,code);
        return ret;
    }
    else{ // any_is_float
        ValueRef* v1 = BuildSitofp(builder, lhs);
        ValueRef* v2 = BuildSitofp(builder, rhs);
        auto ret = new Float_Var();
        string src1 = v1->get_Ref(), src2 = v2->get_Ref(), dest = ret->get_Ref();
        string code = dest + " = fadd float "+ src1 + ", " + src2;

        IRInstruction * instr = new BinaryInstruction(ret,lhs,rhs,ADD);
        builder->AppendCode(instr,code);
        return ret;
    }
}

ValueRef* BuildSub(Builder* builder, ValueRef* lhs, ValueRef* rhs){
    if(Both_Const(lhs,rhs)){
        if(lhs->type == rhs->type && lhs->type == 0){
            int op1 = stoi(lhs->get_Ref());
            int op2 = stoi(rhs->get_Ref());
            return new Int_Const(op1 - op2);
        }
        else{
            float op1 = stof(lhs->get_Ref());
            float op2 = stof(rhs->get_Ref());
            return new Float_Const(op1 - op2);
        }
    }
    else if(Both_Int(lhs,rhs)){
        auto ret = new Int_Var();
        string src1 = lhs->get_Ref(), src2 = rhs->get_Ref(), dest = ret->get_Ref();
        string code = dest + " = sub i32 "+ src1 + ", " + src2;

        IRInstruction * instr = new BinaryInstruction(ret,lhs,rhs,SUB);
        builder->AppendCode(instr,code);
        return ret;
    }
    else{
        ValueRef* v1 = BuildSitofp(builder, lhs);
        ValueRef* v2 = BuildSitofp(builder, rhs);
        auto ret = new Float_Var();
        string src1 = v1->get_Ref(), src2 = v2->get_Ref(), dest = ret->get_Ref();
        string code = dest + " = fsub float "+ src1 + ", " + src2;

        IRInstruction * instr = new BinaryInstruction(ret,lhs,rhs,SUB);
        builder->AppendCode(instr,code);
        return ret;
    }
}

ValueRef* BuildMul(Builder* builder, ValueRef* lhs, ValueRef* rhs){
    if(Both_Const(lhs,rhs)){
        if(lhs->type == rhs->type && lhs->type == 0){
            int op1 = stoi(lhs->get_Ref());
            int op2 = stoi(rhs->get_Ref());
            return new Int_Const(op1 * op2);
        }
        else{
            float op1 = stof(lhs->get_Ref());
            float op2 = stof(rhs->get_Ref());
            return new Float_Const(op1 * op2);
        }
    }
    else if(Both_Int(lhs,rhs)){
        auto ret = new Int_Var();
        string src1 = lhs->get_Ref(), src2 = rhs->get_Ref(), dest = ret->get_Ref();
        string code = dest + " = mul i32 "+ src1 + ", " + src2;

        IRInstruction * instr = new BinaryInstruction(ret,lhs,rhs,MUL);
        builder->AppendCode(instr,code);
        return ret;
    }
    else{
        ValueRef *v1 = BuildSitofp(builder, lhs);
        ValueRef *v2 = BuildSitofp(builder, rhs);
        auto ret = new Float_Var();
        string src1 = v1->get_Ref(), src2 = v2->get_Ref(), dest = ret->get_Ref();
        string code = dest + " = fmul float "+ src1 + ", " + src2;

        IRInstruction * instr = new BinaryInstruction(ret,lhs,rhs,MUL);
        builder->AppendCode(instr,code);
        return ret;
    }
}

ValueRef* BuildDiv(Builder* builder, ValueRef* lhs, ValueRef* rhs){

    if(Both_Const(lhs,rhs)){
        if(lhs->type == rhs->type && lhs->type == 0){
            int op1 = stoi(lhs->get_Ref());
            int op2 = stoi(rhs->get_Ref());
            return new Int_Const(op1 / op2);
        }
        else{
            float op1 = stof(lhs->get_Ref());
            float op2 = stof(rhs->get_Ref());
            return new Float_Const(op1 / op2);
        }
    }
    else if(Both_Int(lhs,rhs)){
        auto ret = new Int_Var();
        string src1 = lhs->get_Ref(), src2 = rhs->get_Ref(), dest = ret->get_Ref();
        string code = dest + " = sdiv i32 "+ src1 + ", " + src2;

        IRInstruction * instr = new BinaryInstruction(ret,lhs,rhs,DIV);
        builder->AppendCode(instr,code);
        return ret;
    }
    else{
        ValueRef* v1 = BuildSitofp(builder, lhs);
        ValueRef* v2 = BuildSitofp(builder, rhs);
        auto ret = new Float_Var();
        string src1 = v1->get_Ref(), src2 = v2->get_Ref(), dest = ret->get_Ref();
        string code = dest + " = fdiv float "+ src1 + ", " + src2;

        IRInstruction * instr = new BinaryInstruction(ret,lhs,rhs,DIV);
        builder->AppendCode(instr,code);
        return ret;
    }
}

ValueRef* BuildMod(Builder* builder, ValueRef* lhs, ValueRef* rhs){
    if(Both_Const(lhs,rhs)){
        if(lhs->type == rhs->type && lhs->type == 0){
            int op1 = stoi(lhs->get_Ref());
            int op2 = stoi(rhs->get_Ref());
            return new Int_Const(op1 % op2);
        }
        else{
            cerr << "invalid operands to binary expression('float' to 'float')" << endl;
            throw exception();
        }
    }
    else if(Both_Int(lhs,rhs)){
        auto ret = new Int_Var();
        string src1 = lhs->get_Ref(), src2 = rhs->get_Ref(), dest = ret->get_Ref();
        string code = dest + " = srem i32 "+ src1 + ", " + src2;

        IRInstruction * instr = new BinaryInstruction(ret,lhs,rhs,MOD);
        builder->AppendCode(instr,code);
        return ret;
    }
    else{
        //TODO:'float'改成getType
        cerr << "invalid operands to binary expression('float' to 'float')" << endl;
        throw exception();
    }
}


ValueRef* BuildLT(Builder* builder, ValueRef* lhs, ValueRef* rhs){
    lhs = BuildZext(builder,lhs); // lhs与rhs可能为bool型，因此先做一轮转化
    rhs = BuildZext(builder,rhs);

    if(Both_Const(lhs,rhs)){
        if(lhs->type == rhs->type && lhs->type == 0){
            int op1 = stoi(lhs->get_Ref());
            int op2 = stoi(rhs->get_Ref());
            ValueRef* res = op1<op2 ? new Int_Const(1) : new Int_Const(0);
            return res;
        }
        else{
            float op1 = stof(lhs->get_Ref());
            float op2 = stof(rhs->get_Ref());
            ValueRef* res = op1<op2 ? new Int_Const(1) : new Int_Const(0);
            return res;
        }
    }
    else if(Both_Int(lhs,rhs)){
        auto result = new Bool_Var();

        string code = result->get_Ref() + " = icmp slt i32 " + lhs->get_Ref() + ", " + rhs->get_Ref(); // cmp
        IRInstruction* instr = new CmpInstruction(result,lhs,rhs,LT);

        builder->AppendCode(instr,code);
        return result;
    }
    else{
        lhs = BuildSitofp(builder,lhs); // int -> float
        rhs = BuildSitofp(builder,rhs);
        auto result = new Bool_Var();

        string code = result->get_Ref() + " = fcmp olt float " + lhs->get_Ref() + ", " + rhs->get_Ref(); // cmp
        IRInstruction* instr = new CmpInstruction(result,lhs,rhs,LT);
        builder->AppendCode(instr,code);

        return result;
    }
}

ValueRef* BuildLE(Builder* builder, ValueRef* lhs, ValueRef* rhs){
    lhs = BuildZext(builder,lhs); // lhs与rhs可能为bool型，因此先做一轮转化
    rhs = BuildZext(builder,rhs);

    if(Both_Const(lhs,rhs)){
        if(lhs->type == rhs->type && lhs->type == 0){
            int op1 = stoi(lhs->get_Ref());
            int op2 = stoi(rhs->get_Ref());
            ValueRef* res = op1<=op2 ? new Int_Const(1) : new Int_Const(0);
            return res;
        }
        else{
            float op1 = stof(lhs->get_Ref());
            float op2 = stof(rhs->get_Ref());
            ValueRef* res = op1<=op2 ? new Int_Const(1) : new Int_Const(0);
            return res;
        }
    }
    else if(Both_Int(lhs,rhs)){
        auto result = new Bool_Var();

        string code = result->get_Ref() + " = icmp sle i32 " + lhs->get_Ref() + ", " + rhs->get_Ref(); // cmp
        IRInstruction* instr = new CmpInstruction(result,lhs,rhs,LE);
        builder->AppendCode(instr,code);

        return result;
    }
    else{
        lhs = BuildSitofp(builder,lhs); // int -> float
        rhs = BuildSitofp(builder,rhs);
        auto result = new Bool_Var();

        string code = result->get_Ref() + " = fcmp ole float " + lhs->get_Ref() + ", " + rhs->get_Ref(); // cmp
        IRInstruction* instr = new CmpInstruction(result,lhs,rhs,LE);
        builder->AppendCode(instr,code);

        return result;
    }
}

ValueRef* BuildEquals(Builder* builder, ValueRef* lhs, ValueRef* rhs){
    lhs = BuildZext(builder,lhs); // lhs与rhs可能为bool型，因此先做一轮转化
    rhs = BuildZext(builder,rhs);

    if(Both_Const(lhs,rhs)){
        if(lhs->type == rhs->type && lhs->type == 0){
            int op1 = stoi(lhs->get_Ref());
            int op2 = stoi(rhs->get_Ref());
            ValueRef* res = op1==op2 ? new Int_Const(1) : new Int_Const(0);
            return res;
        }
        else{
            float op1 = stof(lhs->get_Ref());
            float op2 = stof(rhs->get_Ref());
            ValueRef* res = op1==op2 ? new Int_Const(1) : new Int_Const(0);
            return res;
        }
    }
    else if(Both_Int(lhs,rhs)){
        auto result = new Bool_Var();

        string code = result->get_Ref() + " = icmp eq i32 " + lhs->get_Ref() + ", " + rhs->get_Ref(); // cmp
        IRInstruction* instr = new CmpInstruction(result,lhs,rhs,EQ);
        builder->AppendCode(instr,code);

        return result;
    }
    else{
        lhs = BuildSitofp(builder,lhs); // int -> float
        rhs = BuildSitofp(builder,rhs);
        auto result = new Bool_Var();

        string code = result->get_Ref() + " = fcmp oeq float " + lhs->get_Ref() + ", " + rhs->get_Ref(); // cmp
        IRInstruction* instr = new CmpInstruction(result,lhs,rhs,EQ);
        builder->AppendCode(instr,code);

        return result;
    }
}

ValueRef* BuildUnequal(Builder* builder, ValueRef* lhs, ValueRef* rhs){
    lhs = BuildZext(builder,lhs); // lhs与rhs可能为bool型，因此先做一轮转化
    rhs = BuildZext(builder,rhs);

    if(Both_Const(lhs,rhs)){
        if(lhs->type == rhs->type && lhs->type == 0){
            int op1 = stoi(lhs->get_Ref());
            int op2 = stoi(rhs->get_Ref());
            ValueRef* res = op1!=op2 ? new Int_Const(1) : new Int_Const(0);
            return res;
        }
        else{
            float op1 = stof(lhs->get_Ref());
            float op2 = stof(rhs->get_Ref());
            ValueRef* res = op1!=op2 ? new Int_Const(1) : new Int_Const(0);
            return res;
        }
    }
    else if(Both_Int(lhs,rhs)){
        auto result = new Bool_Var();

        string code = result->get_Ref() + " = icmp ne i32 " + lhs->get_Ref() + ", " + rhs->get_Ref(); // cmp
        IRInstruction* instr = new CmpInstruction(result,lhs,rhs,NE);
        builder->AppendCode(instr,code);

        return result;
    }
    else{
        lhs = BuildSitofp(builder,lhs); // int -> float
        rhs = BuildSitofp(builder,rhs);
        auto result = new Bool_Var();

        string code = result->get_Ref() + " = fcmp une float " + lhs->get_Ref() + ", " + rhs->get_Ref(); // cmp
        IRInstruction* instr = new CmpInstruction(result,lhs,rhs,NE);
        builder->AppendCode(instr,code);

        return result;
    }
}

ValueRef* BuildNeg(Builder* builder, ValueRef* ref){
    switch(ref->type){
        case IntConst:{
            int val = stoi(ref->get_Ref());
            return new Int_Const(-val);
        }
        case FloatConst:{
            auto val = stof(ref->get_Ref());
            return new Float_Const(-val);
        }
        case IntVar:{
            return BuildSub(builder,new Int_Const(0),ref);
        }
        case FloatVar:{
            auto *ret = new Float_Var();
            string code = ret->name + " = fneg float " + ref->get_Ref();

            IRInstruction* instr = new FNegInstruction(ret,ref);
            builder->AppendCode(instr,code);
            return ret;
        }
        default:{
            cerr << "not defined!" << endl;
            throw exception();
        }
    }
}

ValueRef* BuildNot(Builder* builder, ValueRef* ref){
    switch(ref->type){
        case IntConst:{
            int val = stoi(ref->get_Ref());
            return new Int_Const(!val);
        }
        case FloatConst:{
            auto val = stof(ref->get_Ref());
            return new Int_Const(val != 0.0);
        }
        case IntVar:{
            string src_name = ref->get_Ref();
            ValueRef * temp1 = BuildCmp(builder,ref,new Int_Const(0),NE); // icmp ne 0
            ValueRef * temp2 = BuildXorTrue(builder,temp1);
            ValueRef * ret = BuildZext(builder,temp2);
            return ret;
        }
        case FloatVar:{
            string src_name = ref->get_Ref();
            ValueRef * temp1 = BuildCmp(builder,ref,new Float_Const(0.0),NE); //fcmp ne 0.0
            ValueRef * temp2 = BuildXorTrue(builder,temp1);
            ValueRef * ret = BuildZext(builder,temp2);
            return ret;
        }
        default:{
            cerr << "not defined!" << endl;
            throw exception();
        }
    }
}

ValueRef* BuildZext(Builder* builder, ValueRef* ref){
    if(ref->type != BoolVar){
        return ref;
    }
    auto* ret = new Int_Var();
    string code = ret->name + " = zext i1 " + ref->get_Ref() + " to i32";
    IRInstruction* instr = new ZExtInstruction(ret,ref);
    builder->AppendCode(instr,code);

    return ret;
}

ValueRef* BuildSitofp(Builder* builder, ValueRef* ref){
    if(ref->type == FloatConst || ref->type == FloatVar){
        return ref;
    }
    else if(ref->type == IntConst){
        return new Float_Const(stof(ref->get_Ref()));
    }
    else if(ref->type == IntVar){ //IntVar
        auto* ret = new Float_Var();
        string code = ret->get_Ref() + " = sitofp i32 "+ ref->get_Ref() + " to float";   // e.g %6 = sitofp i32 %5 to float
        IRInstruction* instr = new FPSIInstruction(ret,ref);
        builder->AppendCode(instr,code);

        return ret;
    }
    else{
        cerr << "not a numeric type!" << endl;
        throw exception();
    }
}

ValueRef* BuildFPTrunc(Builder* builder, ValueRef* ref){
    if(ref->type == IntConst || ref->type == IntVar){
        return ref;
    }
    else if(ref->type == FloatConst){
        return new Int_Const(stoi(ref->get_Ref()));
    }
    else if(ref->type == FloatVar){ //IntVar
        auto* ret = new Int_Var();
        string code = ret->get_Ref() + " = fptosi float "+ ref->get_Ref() + " to i32";   // e.g %11 = fptosi float %10 to i32
        IRInstruction* instr = new FPSIInstruction(ret,ref);
        builder->AppendCode(instr,code);

        return ret;
    }
    else{
        cerr << "not a numeric type!" << endl;
        throw exception();
    }
}

ValueRef* BuildXorTrue(Builder* builder,ValueRef* ref){
    auto* ret = new Bool_Var();
    string code = ret->name + " = xor i1 " + ref->get_Ref() + ", true";
    IRInstruction* instr = new XorInstruction(ret,ref);
    builder->AppendCode(instr,code);

    return ret;
}

ValueRef* BuildLoad(Builder* builder,Pointer *ptr){ // load ptr指向的位置(用于数组load)
    Type* eleType = ptr->pointerType->elementType;
    switch(eleType->type){
        case INT32TYPE:{
            auto ret = new Int_Var();
            string code = ret->get_Ref() + " = load i32, i32* " + ptr->name + ", align 4"; // (e.g %5 = load i32, i32* %2, align 4)
            IRInstruction* instr = new LoadInstruction(ret,ptr);
            builder->AppendCode(instr,code);

            return ret;
        }
        case FLOATTYPE:{
            auto ret = new Float_Var();
            string code = ret->get_Ref() + " = load float, float* " + ptr->name + ", align 4"; // (e.g %7 = load float, float* %3, align 4)
            IRInstruction* instr = new LoadInstruction(ret,ptr);
            builder->AppendCode(instr,code);

            return ret;
        }
        case POINTERTYPE:{
            PointerType* p = ptr->pointerType;
            Pointer* ret = (Pointer*)(ptr->getElement());
            string code = ret->get_Ref() + " = load " + ret->get_Type() + ", " + p->getTypeName() + ' ' + ptr->name  + ", align 8"; // (e.g %4 = load i32*, i32** %2, align 8)
            IRInstruction* instr = new LoadInstruction(ret,ptr);
            builder->AppendCode(instr,code);

            return ret;
        }
        case ARRAYTYPE:
        case FUNCTYPE:
        case VOIDTYPE: {
            cerr << "Non-numeric Type can't be loaded!" << endl;
            throw exception();
        }
    }
}

ValueRef* BuildLoad(Builder* builder,Symbol *symbol){ // load symbol(符号load)
    Type* eleType = symbol->symbolType;
    switch(eleType->type){
        case INT32TYPE:{
            auto ret = new Int_Var();
            string code = ret->get_Ref() + " = load i32, i32* " + symbol->name + ", align 4"; // (e.g %5 = load i32, i32* %2, align 4)
            IRInstruction* instr = new LoadInstruction(ret,symbol);
            builder->AppendCode(instr,code);

            return ret;
        }
        case FLOATTYPE:{
            auto ret = new Float_Var();
            string code = ret->get_Ref() + " = load float, float* " + symbol->name + ", align 4"; // (e.g %7 = load float, float* %3, align 4)
            IRInstruction* instr = new LoadInstruction(ret,symbol);
            builder->AppendCode(instr,code);

            return ret;
        }
        case POINTERTYPE:{
            Pointer* ret = new Pointer((PointerType*) eleType);
            string code = ret->get_Ref() + " = load " + ret->get_Type() + ", " + eleType->getTypeName() + "* " + symbol->name  + ", align 8"; // (e.g %4 = load i32*, i32** %2, align 8)
            IRInstruction* instr = new LoadInstruction(ret,symbol);
            builder->AppendCode(instr,code);
            return ret;
        }
        case ARRAYTYPE:
        case FUNCTYPE:
        case VOIDTYPE: {
            cerr << "Non-numeric Type can't be loaded!" << endl;
            throw exception();
        }
    }
}

void BuildStore(Builder* builder,Pointer* ptr,ValueRef* ref){ // save ref -> ptr指向的位置(数组用)
    PointerType *p = ptr->pointerType;
    if(p->elementType->type == INT32TYPE){
        ref = BuildFPTrunc(builder,ref);
    }
    else if(p->elementType->type == FLOATTYPE){
        ref = BuildSitofp(builder,ref);
    }

    string code = "store " + p->elementType->getTypeName() + ' ' + ref->get_Ref() + ", " + p->getTypeName() + ' ' + ptr->name + ", align 8"; // (e.g store i32* %0, i32** %2, align 8)
    IRInstruction* instr = new StoreInstruction(ptr,ref);
    builder->AppendCode(instr,code);
}

void BuildStore(Builder* builder,Symbol* symbol,ValueRef* ref){ // save ref -> symbol(符号用)
    Type* eleType = symbol->symbolType;
    if(eleType->type == INT32TYPE){
        ref = BuildFPTrunc(builder,ref);
    }
    else if(eleType->type == FLOATTYPE){
        ref = BuildSitofp(builder,ref);
    }

    string code = "store " + eleType->getTypeName() + ' ' + ref->get_Ref() + ", " + eleType->getTypeName() + "* " + symbol->name + ", align 8"; // (e.g store i32* %0, i32** %2, align 8)
    IRInstruction* instr = new StoreInstruction(symbol,ref);
    builder->AppendCode(instr,code);
}


Symbol* BuildAlloc(Builder* builder,Type* type,const std::string& name,bool is_const) { // 局部变量
    switch (type->type) {
        case VOIDTYPE: {
            break;
        }
        case INT32TYPE: {
            string varname = Utils::getNewLabel(name);
            Symbol* ret = new Symbol(type,varname,is_const);

            string code = varname + " = alloca i32, align 4"; // e,g %a = alloca i32, align 4
            IRInstruction* instr = new AllocaInstruction(ret,type);
            builder->AppendCode(instr,code);

            return ret;
        }
        case FLOATTYPE: {
            string varname = Utils::getNewLabel(name);
            Symbol* ret = new Symbol(type,varname,is_const);

            string code = varname + " = alloca float, align 4"; // e.g %a = alloca float, align 4
            IRInstruction* instr = new AllocaInstruction(ret,type);
            builder->AppendCode(instr,code);

            return ret;
        }
        case ARRAYTYPE: {
            string varname = Utils::getNewLabel(name);
            Symbol* ret = new Symbol(type,varname,is_const);

            string code = varname + " = alloca " + type->getTypeName() + ", align 16"; // e.g %2 = alloca [2 x [2 x i32]], align 16
            IRInstruction* instr = new AllocaInstruction(ret,type);
            builder->AppendCode(instr,code);

            return ret;
        }
        case POINTERTYPE:{
            string varname = Utils::getNewLabel(name);
            Symbol* ret = new Symbol(type,varname,is_const);

            string code = varname + " = alloca " + type->getTypeName() + ", align 8"; // e.g %2 = alloca i32*, align 8
            IRInstruction* instr = new AllocaInstruction(ret,type);
            builder->AppendCode(instr,code);

            return ret;
        }
        default: {
            break;
        }
    }
    return nullptr;
}

Symbol* AddGlobal(Builder* builder,Type* type,const std::string& name,ValueRef *initVal,bool is_const) { // 全局变量
    Symbol * symbol = new Symbol(type,name,is_const);
    if(initVal == nullptr){
        string code;
        switch(type->type){
            case INT32TYPE: code = name + " = dso_local global i32 0, align 4"; break;
            case FLOATTYPE: code = name + " = dso_local global float 0.000000e+00, align 4"; break;
            case ARRAYTYPE: code = name + " = dso_local global " + zeroInitializer(type) + ", align 16"; break;
            default: throw exception();
        }

        IRInstruction* instr = new AddGlobalInstruction(symbol,type,initVal);
        builder->AppendCode(instr,code);

        if(is_const) {
            switch (symbol->symbolType->type) {
                case INT32TYPE: {
                    symbol->constVal = new Int_Const(0);
                    break;
                }
                case FLOATTYPE: {
                    symbol->constVal = new Float_Const(0.0);
                    break;
                }
                case ARRAYTYPE:{
                    vector<any> v;
                    symbol->constVal = new Array_Const(v);
                }
                default: throw exception();
            }
        }
    }
    else{ // has initVal
        if(type->type != ARRAYTYPE) {
            if (initVal->type != IntConst && initVal->type != FloatConst) {
                cerr << "initializer element is not a compile-time constant" << endl;
                throw exception();
            }
            string str = initVal->type == IntConst ? "global" : "constant";
            string code = name + " = dso_local "+str+" " + type->getTypeName() + ' ' + initVal->get_Ref() + ", align 4";
            //string code = name + " = dso_local global " + type->getTypeName() + ' ' + initVal->get_Ref() + ", align 4";

            IRInstruction* instr = new AddGlobalInstruction(symbol,type,initVal);
            builder->AppendCode(instr,code);
            if (is_const) {
                switch (symbol->symbolType->type) {
                    case INT32TYPE: {
                        symbol->constVal = new Int_Const(stoi(initVal->get_Ref()));
                        break;
                    }
                    case FLOATTYPE: {
                        symbol->constVal = new Float_Const(stof(initVal->get_Ref()));
                        break;
                    }
                    default:
                        break;
                }
            }
        }
        else{ //数组单独考虑
            symbol->constVal = initVal;
            string code = name + " = dso_local global " + nonZeroInitializer(((Array_Const*)initVal)->values ,type) + ", align 16";
            IRInstruction* instr = new AddGlobalInstruction(symbol,type,initVal);
            builder->AppendCode(instr,code);
        }
    }
    return symbol;
}

string zeroInitializer(Type *type) {
    if(type->type == ARRAYTYPE){
        return type->getTypeName() + " zeroinitializer";
    }
    else{
        return type->type == INT32TYPE? "i32 0":"float 0.000000e+00";
    }
    /*
    string ret = "";
    ret += type->getTypeName() + ' ';
    if(type->type == ARRAYTYPE){
        ret += '[';
        string code = zeroInitializer(((ArrayType*) type)->elementType);
        for(int i=0;i<((ArrayType*) type)->ele_cnt;i++){
            if(i != 0) ret += ", ";
            ret += code;
        }
        ret += ']';
        return ret;
    }
    else{
        return type->type == INT32TYPE? "i32 0":"float 0.000000e+00";
    }
    */
}


string nonZeroInitializer(std::any values,Type* type){
    string ret = "";
    ret += type->getTypeName() + ' ';
    if(type->type == ARRAYTYPE){
        ret += '[';

        auto v = any_cast<vector<std::any>>(values);
        int size = v.size();

        for(int i=0;i<((ArrayType*) type)->ele_cnt;i++){
            if(i != 0) ret += ", ";
            if(i < size)
                ret += nonZeroInitializer(v[i],((ArrayType*) type)->elementType);
            else
                ret += zeroInitializer(((ArrayType*) type)->elementType);
        }
        ret += ']';
        return ret;
    }
    else{
        auto initVal = any_cast<ValueRef*>(values);
        return initVal->get_Type() + ' ' + initVal->get_Ref();
    }
}


ValueRef* BuildCmp(Builder* builder,ValueRef* lhs, ValueRef* rhs,cmpType type){
    //EQ,NE,UGT,UGE,ULT,ULE,SGT,SGE,SLT,SLE
    switch(type){
        case EQ: return BuildEquals(builder,lhs,rhs);
        case NE: return BuildUnequal(builder,lhs,rhs);
        case GT: return BuildLT(builder,rhs,lhs);
        case GE: return BuildLE(builder,rhs,lhs);
        case LT: return BuildLT(builder,lhs,rhs);
        case LE: return BuildLE(builder,lhs,rhs);
    }
}

ValueRef *BuildCall(Builder *builder,Function *func, std::vector<ValueRef *>& args) {
    //e.g %0 = call i32 @func(i32 %1) -> normal call
    //    call void @foo(i32 %2)      -> void call
    string command;
    ValueRef* ret;
    switch(func->funcType->retType->type){
        case VOIDTYPE:{
            command = "call void @";
            break;
        }
        case INT32TYPE:{
            ret = new Int_Var();
            command = ret->get_Ref() + " = call i32 @";
            break;
        }
        case FLOATTYPE:{
            ret = new Float_Var();
            command = ret->get_Ref() + " = call float @";
            break;
        }
        default: {
            cerr << "Return type is not basic type!!!" << endl;
            throw exception();
        }
    }

    command += func->name;
    command += '(';
    // add params
    bool not_first = false;
    for(ValueRef * ref: args){
        if(not_first) command += ',';
        not_first = true;
        command += ref->get_Type() + ' ' + ref->get_Ref();
    }
    command += ')';

    IRInstruction* instr = new CallInstruction(ret,func,args);
    builder->AppendCode(instr,command);
    return ret;
}

void BuildBr(Builder *builder, BasicBlock *block) {
    string code = "br label %" + block->label;
    builder->LinkBlock(block);

    IRInstruction* instr = new BrInstruction(block);
    builder->AppendCode(instr,code);
}

void BuildCondBr(Builder *builder, ValueRef *cond, BasicBlock *ifTrue, BasicBlock *ifFalse) {
    // e.g br i1 %8, label %9, label %12
    string code = "br i1 " + cond->get_Ref() + ", label %" + ifTrue->label + ", label %" + ifFalse->label;

    builder->LinkBlock(ifTrue);
    builder->LinkBlock(ifFalse);

    IRInstruction* instr = new CondBrInstruction(cond,ifTrue,ifFalse);
    builder->AppendCode(instr,code);
}

Pointer *BuildGEP(Builder* builder,ValueRef *ptr, std::vector<ValueRef *>& idxs) {
    // GEP返回值只可能是Pointer,因此Array在最后要多做一次IndexOfArray以转化为Pointer

    ValueRef *temp = ptr;
    for(ValueRef * i : idxs){
        temp = IndexOfArray(builder,temp,i);
    }
    if(temp->type == Ptr){
        return (Pointer*)temp;
    }
    else { // temp->type == Arr
        // GEP返回值只可能是Pointer,因此Array在最后要多做一次IndexOfArray以转化为Pointer
        ValueRef *ret = IndexOfArray(builder,temp,new Int_Const(0));
        if(ret->type == Ptr){
            return (Pointer*)ret;
        }
        else{ // 返回值仍为Array,直接用构造函数转化类型
            return new Pointer(ret);
        }
    }
}

ValueRef *IndexOfArray(Builder* builder,ValueRef *ref,ValueRef * index){
    Type *ptrType; // 只有可能是指针类型或数组类型
    if(ref->type == Arr){
        ptrType = ((Array*)ref)->arrayType;
    }
    else if(ref->type == Ptr){
        ptrType = ((Pointer*)ref)->pointerType;
    }
    else{ // ref->type == Symbol
        ptrType = ((Symbol*)ref)->symbolType;
    }

    ValueRef *ret = ref->getElementPtr();
    if(ptrType->type == ARRAYTYPE){
        //数组的GEP: 数组类型* -> 数组类型
        string code = ret->get_Ref() + " = getelementptr " + ref->get_Type() + ", " + ref->get_Type() + "* " + ref->get_Ref() + ", i32 0, i32 " + index->get_Ref();
        //e.g %4 = getelementptr inbounds [2 x [2 x i32]], [2 x [2 x i32]]* %2, i64 0, i64 1

        IRInstruction* instr = new GEPInstruction(ret,ref,index);
        builder->AppendCode(instr,code);
    }
    else if(ptrType->type == POINTERTYPE){
        //指针的GEP: 元素类型* -> 元素类型
        string eleName = ((PointerType*)ptrType)->elementType->getTypeName();
        string code = ret->get_Ref() + " = getelementptr " + eleName + ", " + eleName + "* " + ref->get_Ref() + ", i32 " + index->get_Ref();
        //e.g %4 = getelementptr inbounds [2 x [2 x i32]], [2 x [2 x i32]]* %2, i64 0

        IRInstruction* instr = new GEPInstruction(ret,ref,index);
        builder->AppendCode(instr,code);
    }
    else{
        cerr << "Not a pointer/arr type!" << endl;
        throw exception();
    }
    return ret;
}

void BuildRet(Builder *builder, ValueRef *retVal) {
    string code;
    if (retVal == nullptr)
        code = "ret void";
    else {
        code = "ret " + retVal->get_Type() + ' ' + retVal->get_Ref();
    }

    IRInstruction* instr = new RetInstruction(retVal);
    builder->AppendCode(instr,code);
}

bool Both_Const(ValueRef* lhs, ValueRef* rhs){
    return lhs->type <= FloatConst && rhs->type <= FloatConst;
}

bool Both_Int(ValueRef* lhs, ValueRef* rhs){
    return (lhs->type == IntConst || lhs->type == IntVar) && (rhs->type == IntConst || rhs->type == IntVar);
}
