#include "codegen/LiveAnalysis.hpp"
#include "codegen/MachineBlock.hpp"
#include "codegen/MachineOperand.hpp"
#include "codegen/MachineUnit.hpp"
#include "BasicBlock.h"

#include <algorithm>

void LiveAnalysis::pass(MachineUnit *unit) {
    for (auto func : unit->func_list) {
        calcLiveDefUse(func);
        calcLiveInOut(func);
    }
}

void LiveAnalysis::calcLiveDefUse(MachineFunction *mFunc) {
    for (auto mBlock : mFunc->block_list) {
        for (auto inst : mBlock->inst_list) {
            auto &live_use = mBlock->live_use;
            auto &live_def = mBlock->live_def;

            for (auto use : inst->use_list) {
                if (use->type != MachineOperand::VREG) continue;
                const auto &vreg = use->vreg_name;
                if (live_def.find(vreg) == live_def.end()) {
                    live_use.insert(vreg);
                }
            }
            
            for (auto def : inst->def_list) {
                if (def->type != MachineOperand::VREG) continue;
                const auto &vreg = def->vreg_name;
                live_def.insert(vreg);
            }
        }
    }
}

void LiveAnalysis::calcLiveInOut(MachineFunction *mFunc) {
    bool change;
    do {
        change = false;
        // 逆序遍历mFunc->block_list(当前函数的所有block)
        for (auto it = mFunc->block_list.rbegin(); it != mFunc->block_list.rend(); ++it) {
            MachineBlock *mBlock = *it;
            auto old_live_out(mBlock->live_out);
            mBlock->live_out.clear();
            for (auto bb : mBlock->bb->succ) {
                auto succ = bb->mBlock;
                mBlock->live_out.insert(succ->live_in.begin(), succ->live_in.end());
                auto tmp(mBlock->live_use);

                // 求差集和并集
                std::set_difference(mBlock->live_out.begin(), mBlock->live_out.end(),
                                    mBlock->live_def.begin(), mBlock->live_def.end(),
                                    std::inserter(tmp, tmp.begin()));
                mBlock->live_in = std::move(tmp);
            }
            if (old_live_out != mBlock->live_out) change = true;
        }
    } while (change);
}