#include <iostream>
#include <ostream>
#include <string>
#include "antlr4-common.h"
#include "SysYLexer.h"
#include "SysYParser.h"
#include "MyVisitor.h"
#include "codegen/AsmBuilder.hpp"
#include "codegen/LinearScan.hpp"
#include "codegen/LiveAnalysis.hpp"
#include "codegen/MachineCodePass.hpp"
#include "codegen/MachineFunction.hpp"

#include "HIR-opt/HOptimizer.h"

#ifndef TESTCASE_PATH
#define TESTCASE_PATH "./testcases"
#endif

int main(int argc, const char* argv[]) {
    std::string testcase_path(TESTCASE_PATH);
    string test_file;
    std::string output_file;
    if(argc == 1){
         test_file = testcase_path + "/test.sy";
    }
    else if(argc == 2) {// local test
        test_file = testcase_path + "/" + argv[1];
    }
    else{ // argc > 2 : online test
        test_file = argv[4];
        output_file = argv[3];
    }


    std::ifstream stream(test_file);
    assert(stream);
    antlr4::ANTLRInputStream input(stream);
    antlr4::SysYLexer lexer(&input);
    antlr4::CommonTokenStream tokens(&lexer);
    antlr4::SysYParser parser(&tokens);

    // gen HIR
    antlr4::tree::ParseTree *tree = parser.program();
    MyVisitor* visitor = new MyVisitor();
    visitor->add_sylib_func();
    visitor->visit(tree);

    // opt HIR
    GlobalUnit *gu = visitor->globalUnit;
    HOptimizer* opt = new HOptimizer(gu);
    //opt->Optimize();

    gu->Emit();

    auto builder = new AsmBuilder();
    auto mUnit = new MachineUnit();
    visitor->globalUnit->codegen(mUnit, builder);
    mUnit->output(std::cout);

    std::vector<MachineCodePass *> mPasses;
    mPasses.push_back(new LiveAnalysis());
    mPasses.push_back(new LinearScan());
    for (auto mPass : mPasses) {
        mPass->pass(mUnit);
    }
    if (output_file.empty()) output_file = std::string(TESTCASE_PATH) + "/" + "out.s";
    std::ofstream output_stream(output_file, std::ios::out | std::ios::trunc);
    mUnit->output(output_stream);
    // mUnit->output(std::cerr);

    return 0;
}
